/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import * as request from "supertest";
import { expect } from "chai";
import * as server from "../../main";
import { dataCtrlScenario as tests } from "../../../test/scenario";
import { Query } from "../../common/query";

interface StrQuery {
    metrics: string;
    dimensions: string;
    filters?: string;
    sort?: string;
}

function parseQuery(obj: Query): StrQuery {
    let r: StrQuery = {
        metrics: "",
        dimensions: ""
    };

    r.metrics = obj.metrics.map ((item) => item.name).join(",");
    r.dimensions = obj.dimensions.map ((item) => item.name).join(",");
    return r;
}

describe("API data controller", () => {

    it("should respond 500 when query has inexistent metric", (done) => {
        request(server)
            .get("/v1/data")
            .query(parseQuery(tests.wrongMet))
            .expect(500)
            .expect((res: any) => {
                const message = "Query execution failed: " +
                "Could not construct query with the given parameters.";
                const error = "The metric named met:this:is:just:a:test" +
                " was not found";
                expect(res.body).to.be.an("object");
                expect(res.body).to.have.property("message");
                expect(res.body).to.have.property("error");
                expect(res.body.message).to.be.eql(message);
                expect(res.body.error).to.be.eql(error);
            })
            .end(done);
    });

    it("should respond 500 when query has inexistent dimension", (done) => {
        request(server)
            .get("/v1/data")
            .query(parseQuery(tests.wrongDim))
            .expect(500)
            .expect((res: any) => {
                const message = "Query execution failed: " +
                "Could not construct query with the given parameters.";
                const error = "The dimension named dim:this:is:just:a:test" +
                " was not found";
                expect(res.body).to.be.an("object");
                expect(res.body).to.have.property("message");
                expect(res.body).to.have.property("error");
                expect(res.body.message).to.be.eql(message);
                expect(res.body.error).to.be.eql(error);
            })
            .end(done);
    });

    it("should respond 500 when query has sort item that is not in query data", (done) => {
        let query = parseQuery(tests.clausal);
        query.sort = "dim:does:not:exist";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(500)
            .expect((res: any) => {
                const message = "Query execution failed: " +
                "Could not construct query with the given parameters.";
                const error = "The item 'dim:does:not:exist'" +
                    " is not present in neither metrics nor dimensions list";
                expect(res.body).to.be.an("object");
                expect(res.body).to.have.property("message");
                expect(res.body).to.have.property("error");
                expect(res.body.message).to.be.eql(message);
                expect(res.body.error).to.be.eql(error);
            })
            .end(done);
    });

    it("should respond 200 and get some data", (done) => {
        request(server)
            .get("/v1/data")
            .query(parseQuery(tests.correct))
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.correct.metrics.map((item) => item.name));
                keys = keys.concat(tests.correct.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });

    it("should respond 200 and get some data, using a single filter", (done) => {
        let query = parseQuery(tests.clausal);
        query.filters = "dim:product:name==Bacon";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(1);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });

    it("should respond 200 and get some data, using filters with OR", (done) => {
        let query = parseQuery(tests.clausal);
        query.filters = "dim:product:name==Trento,dim:product:name==Meat";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(2);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });

    it("should respond 200 and get some data, using filters with AND", (done) => {
        let query = parseQuery(tests.clausal);
        query.filters = "dim:product:name!=Bacon;dim:product:validity!=2018-05-10";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(3);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });
    it("should respond 200 and get some data, sorted", (done) => {
        let query = parseQuery(tests.clausal);
        query.sort = "dim:product:id,met:product:avg:pricein";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                for (let i = 0; i < result.length; ++i) {
                    const row = result[i];
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                    // the magic number will be the first dim:product:id
                    expect(row["dim:product:id"]).to.be.eql(i + 53);
                }
            })
            .end(done);
    });
    it("should respond 200 and get some data, using filters with GREATER", (done) => {
            let query = parseQuery(tests.clausal);
            query.filters = "dim:product:id>55";
            request(server)
                .get("/v1/data")
                .query(query)
                .expect(200)
                .expect((res: any) => {
                    let result = res.body;
                    expect(result).to.be.an("array");
                    expect(result).to.have.length(2);
                    expect(result[0]).to.be.an("object");
                    let keys: string[] = [];
                    keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                    keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                    result.forEach((row: any) => {
                        expect(row).to.be.an("object");
                        expect(row).to.have.all.keys(keys);
                    });
                })
            .end(done);
    });   
    it("should respond 200 and get some data, using filters with LOWEREQ", (done) => {
        let query = parseQuery(tests.clausal);
        query.filters = "dim:product:id<=55";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(3);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.clausal.metrics.map((item) => item.name));
                keys = keys.concat(tests.clausal.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });  
    it("should respond 200 and get some data, using filters with default", (done) => {
        let query = parseQuery(tests.seller);
        query.filters = "dim:sell:registered==false";
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(1);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.seller.metrics.map((item) => item.name));
                keys = keys.concat(tests.seller.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });

    it("should respond 200 and get some data, using filters with default", (done) => {
        let query = parseQuery(tests.expensive);
        query.sort = "met:product:max:pricein"
        request(server)
            .get("/v1/data")
            .query(query)
            .expect(200)
            .expect((res: any) => {
                let result = res.body;
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(tests.expensive.metrics.map((item) => item.name));
                keys = keys.concat(tests.expensive.dimensions.map((item) => item.name));
                result.forEach((row: any) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
            })
            .end(done);
    });

});
