/*
 * Copyright (C) 2016 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Dimension, DimensionStrOptions } from "./dimension";
import { Metric, MetricStrOptions } from "./metric";
import { Clause } from "./clause";
import { Filter } from "./filter";
import { View } from "./view";
import { Query, QueryOpts } from "../common/query";
import { Graph } from "../util/graph";
import { EnumType, EnumTypeOptions} from "./enumType";
import { Source , SourceStrOptions } from "./source";
import { ViewHandler } from "../util/viewHandler";
import { ParsedConfig } from "../util/configParser";

/**
 * Represents the database schema from the BlenDB perspective.
 * The engine is an abstraction of a generic analytics database.
 * All the operations that would be realised in the database
 * are made through the engine. The main purpose of the engine
 * is to translate a localition less query into a localized query.
 * In other words, discover in which view a metric/dimension is.
 */
export class Engine {
    /** Set of views available in the database */
    private views: View[];
    /** Set of metrics available in the database. */
    private metrics: Metric[];
    /** Set of enumerable types available in the database. */
    private enumTypes: EnumType[];
    /** Set of dimensions available in the database. */
    private dimensions: Dimension[];
    /** Set of sources available in the database. */
    private sources: Source[];
    /** Graph which represents the database schema. */
    private graph: Graph;

    /** Constructs an empty database schema. */
    constructor (config: ParsedConfig) {

        this.graph = new Graph();

        this.enumTypes = [];
        this.views = [];
        this.metrics = [];
        this.dimensions = [];
        this.sources = [];

        config.metrics.forEach ((met) => this.addMetric(met));
        config.dimensions.forEach ((dim) => this.addDimension(dim));
        config.views.forEach ((view) => this.addView(view));
        config.enumTypes.forEach ((enumt) => this.addEnumType(enumt));
        config.sources.forEach ((sourc) => this.addSource(sourc));

    }

    /** Gets all the available */
    public getViews(): View[] {
        return this.views;
    }

    /** Gets a string description for all the available metrics. */
    public getMetricsDescription(): MetricStrOptions[] {
       return this.metrics.map((i) => i.strOptions());
    }

    /** Gets a string description for all the available enumerable types. */
    public getEnumTypesDescription(): EnumTypeOptions[] {
        return this.enumTypes.map((i) => i.strOptions());
    }

    /** Gets a string description for all the available sources. */
    public getSourcesDescription(): SourceStrOptions[] {
        return this.sources.map((i) => i.strOptions());
    }

    /** Gets a string description for all the available dimensions. */
    public getDimensionsDescription(): DimensionStrOptions[] {
       return this.dimensions.map((i) => i.strOptions());
    }

    /**
     * Adds a new view to the database schema (engine).
     * @param view - View to be added.
     */
    public addView(view: View): View {
        if (this.graph.addView(view)) {
            this.views.push(view);
            return view;
        }

        return null;
    }

    /**
     * Adds a new enumerable type to the database schema (engine).
     * @param enumType - Enumerable type to be added.
     */
    public addEnumType(enumType: EnumType): EnumType {
        this.enumTypes.push(enumType);
        return enumType;
    }

    /**
     * Adds a new source to the database schema (engine).
     * @param source - Source to be added.
     */
    public addSource(source: Source): Source {
        this.sources.push(source);
        return source;
    }

    /**
     * Adds a new metric to the database schema (engine).
     * @param metric - Metric to be added.
     */
    public addMetric(metric: Metric): Metric {
        if (this.graph.addMetric(metric)) {
            this.metrics.push(metric);
            return metric;
        }

        return null;
    }

    /**
     * Gets the metric object given its name.
     * @param name - Metric's name.
     */
    public getMetricByName(name: string): Metric {
        let result = this.metrics.find(metric => metric.name === name);

        if (!result) {
            throw new Error("The metric named " + name + " was not found");
        }

        return result;
    }

    /**
     * Gets the enumerable type object given its name.
     * @param name - Enumerable type name.
     */
    public getEnumTypeByName(name: string): EnumType {
        let result = this.enumTypes.find(EnumType => EnumType.name === name);

        if (!result) {
            throw new Error("The dataType named '" + name + "' was not found");
        }

        return result;
    }

    /**
     * Gets the source object given its name.
     * @param name - Source's name.
     */
    public getSourceByName(name: string): Source {
        let result = this.sources.find(source => source.name === name);

        if (!result) {
            throw new Error("The source named '" + name + "' was not found");
        }

        return result;
    }

    /**
     * Adds a new dimension to the database schema (engine).
     * @param dimension - Dimension to be added.
     */
    public addDimension(dimension: Dimension): Dimension {
        if (this.graph.addDimension(dimension)) {
            this.dimensions.push(dimension);
            return dimension;
        }

        return null;
    }

    /**
     * Gets the dimension object given its name.
     * @param name - Dimension's name.
     */
    public getDimensionByName(name: string): Dimension {
        let result = this.dimensions.find(dimension => dimension.name === name);

        if (!result) {
            throw new Error("The dimension named " + name + " was not found");
        }

        return result;
    }

    /**
     * Translate a string clause into a clause object.
     * @param strClause - Clause in string format.
     */
    public parseClause(strClause: string): Clause {
        let strFilters = strClause.split(",").filter((item: string) => item !== "");
        let filters: Filter[] = [];
        for (let i = 0; i < strFilters.length; ++i) {
            filters.push(this.parseFilter(strFilters[i]));
        }

        return new Clause({filters: filters});
    }

    /**
     * Translate a string filter into a clause object.
     * @param strFilter - Filter in string format.
     */
    public parseFilter(strFilter: string): Filter {
        let segment = Filter.segment(strFilter);
        if (segment) {
            // Segment never returns NONE
            let op = Filter.parseOperator(segment.operator);
            let target: Metric|Dimension = null;
            try {
                target = this.getDimensionByName(segment.target);
            }

            catch (e) {
                try {
                    target = this.getMetricByName(segment.target);
                }

                catch (e) {
                    target = null;
                }
            }

            if (!target) {
                throw new Error("Filter could not be created: \"" + segment.target + "\" was not found");
            }

            const filter = new Filter({
                target: target,
                operator: op,
                value: segment.value
            });

            if (!filter.isValid) {
                throw new Error("Filter could not be created: Operator \"" + segment.operator + "\" is invalid for target \"" + segment.target + "\"");
            }

            return filter;
        }
        else {
            throw new Error("Filter could not be created: Operator on \"" + strFilter + "\" could not be extracted");
        }
    }

    /**
     * Transform a query (location less) into a view (localized).
     * The main difference between a query object and a view object
     * is that a view object can be built from other view objects
     * and construct a valid query to conventional SGBD's.
     * The query object do not have any location, so can not be
     * translate in a valid SGBD query. In other words this method
     * find a location (FROM clause in SQL) to the query object.
     * @param q - Query to be located.
     */
    public query (q: Query): View {
        return this.selectOptimalView(q);
    }

    /**
     * Selects the best location (view) to a given query.
     * @param q - Query to be located.
     */
    private selectOptimalView (q: Query): View {
        let queries: Query[] = [];
        let qOpt: QueryOpts;
        if (q.metrics.length > 0) {
            for (let i = 0; i < q.metrics.length; ++i) {
                qOpt = { metrics: [q.metrics[i]],
                    dimensions: q.dimensions,
                    clauses: q.clauses,
                    sort: q.sort };
                queries.push(new Query(qOpt));
            }
            const views = queries.map((query) => {
                return ViewHandler.growView(query, this.getCover(query));
            });

            return ViewHandler.queryJoin(q, views);
        }

        else {
            let query = new Query(q);
            return ViewHandler.growView(query, this.getCover(query));
        }

    }

    /**
     * Finds a set of views that has 2 properties:
     * 1- Contain all the attributes (metrics/dimensions)
     * required do make the query.
     * 2- The set of views is connected. This means that
     * if the set has more than one view is possible to take
     * any view in the set exists another view that share
     * at least one dimension. This meke possible to join
     * views.
     * @param q - Query to be covered.
     */
    private getCover (q: Query): View[] {
        const optimalViews = this.graph.cover(q);
        if (optimalViews.length === 0) {
            throw new Error ("Engine views cannot cover the query");
        }

        let matViews: View[] = optimalViews.sort((a, b) => (a.id < b.id) ? -1 : 1);

        let noRepeat: View[] = [matViews[0]];
        for (let i = 1; i < matViews.length; ++i) {
            if (matViews[i - 1].id !== matViews[i].id) {
                noRepeat.push(matViews[i]);
            }
        }

        return noRepeat;

    }
}
