/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { expect } from "chai";
import { ConfigParser, ViewParsingOptions } from "./configParser";
import { Dimension, DimensionStrOptions } from "../core/dimension";
import { RelationType , DataType} from "../common/types";
import { EnumType } from "../core/enumType";
import { EnumHandler } from "../util/enumHandler";
import { MetricStrOptions } from "../core/metric";
import { SourceStrOptions } from "../core/source";

function strToRelationType (str: string): RelationType {
    switch (str) {
        case "day":
            return RelationType.DAY;
        case "month":
            return RelationType.MONTH;
        case "year":
            return RelationType.YEAR;
        default:
            return RelationType.NONE;
    }
}

describe("configParser utility library", () => {
    it("should throw expection for inexistent metric", () => {
        let opts: ViewParsingOptions =  {
            alias: "Test",
            data: "test",
            dimensions: [],
            metrics: ["met:-1"]
        };

        let error: boolean = false;
        try {
            ConfigParser.parseViewOpt(opts, {}, {});
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] Non exist metric set to view " + opts.alias);
        }

        expect(error).to.be.true;
    });

    it("should throw expection for inexistent dimension", () => {
        let opts: ViewParsingOptions =  {
            alias: "Test",
            data: "test",
            dimensions: ["dim:-1"],
            metrics: []
        };

        let error: boolean = false;
        try {
            ConfigParser.parseViewOpt(opts, {}, {});
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] Non exist dimension set to view " + opts.alias);
        }

        expect(error).to.be.true;
    });

    it("should throw expection for inexistent parent for subdimension", () => {
        let opts: DimensionStrOptions =  {
            name: "dim:-1",
            dataType: "integer",
            parent: "dim:0",
            relation: "none"
        };

        let dims: Dimension[] = [];

        let error: boolean = false;
        try {
            ConfigParser.parseDimOpts(opts, dims, null);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] Parent for subdimension " + opts.name + "  not found");
        }

        expect(error).to.be.true;
    });

    it("should throw expection for bad formed clause, invalid operator", () => {
        let opts: ViewParsingOptions =  {
            alias: "thisMUSTgoesWrong",
            data: "someRandomPath",
            origin: true,
            dimensions: ["dim:0"],
            metrics: [],
            clauses: ["dim:0?=1"]
        };

        let dimMap: {[key: string]: Dimension} = {
            "dim:0" : new Dimension({name: "dim:0", dataType: DataType.INTEGER})
        };

        let error: boolean = false;
        try {
            ConfigParser.parseViewOpt(opts, {}, dimMap);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] Bad formed clause: on view \"" + opts.alias +
                   "\" the clause \"dim:0?=1\" could not be created");
        }

        expect(error).to.be.true;
    });

    it("should throw expection for bad formed clause, invalid target", () => {
        let opts: ViewParsingOptions =  {
            alias: "thisMUSTgoesWrong",
            data: "someRandomPath",
            origin: true,
            dimensions: ["dim:0"],
            metrics: [],
            clauses: ["dim:1==1"]
        };

        let dimMap: {[key: string]: Dimension} = {
            "dim:0" : new Dimension({name: "dim:0", dataType: DataType.INTEGER})
        };

        let error: boolean = false;
        try {
            ConfigParser.parseViewOpt(opts, {}, dimMap);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] Bad formed clause: on view \"" + opts.alias +
                   "\" the clause \"dim:1==1\" could not be created");
        }

        expect(error).to.be.true;
    });

    it("should parse subdimentions relations correctly", () => {
        let opts: DimensionStrOptions[] =  [
            {
                name: "dim:day",
                dataType: "integer",
                parent: "dim:0",
                relation: "day"
            },
            {
                name: "dim:month",
                dataType: "integer",
                parent: "dim:0",
                relation: "month"
            },
            {
                name: "dim:year",
                dataType: "integer",
                parent: "dim:0",
                relation: "year"
            },
            {
                name: "dim:none",
                dataType: "integer",
                parent: "dim:0",
                relation: "none"
            }
        ];

        let dims: Dimension[] = [
            new Dimension({name: "dim:a", dataType: DataType.DATE}),
            new Dimension({name: "dim:0", dataType: DataType.DATE})
        ];

        for (let i = 0; i < opts.length; ++i) {
            let parsed = ConfigParser.parseDimOpts(opts[i], dims, null);
            expect(parsed.name).to.be.equal(opts[i].name);
            expect(EnumHandler.stringfyDataType(parsed.dataType)).to.be.equal(opts[i].dataType);
            expect(parsed.parent).to.be.equal(dims[1]);
            expect(parsed.relation).to.be.equal(strToRelationType(opts[i].relation));
        }
    });

    it("should parse correctly enumType with dimension", () => {

        let opts: DimensionStrOptions =  {
            name: "dim:day",
            dataType: "enumtype",
            enumType: "enumtype:5",
            parent: "dim:0",
            relation: "day"
        };
        let dims: Dimension[] = [
            new Dimension({name: "dim:0", dataType: DataType.DATE})
        ];
        let enumMap: {[key: string]: EnumType} = {
            "enumtype:5" : new EnumType({name: "enumtype:5", values: ["nope", "test"]})
        };
        let parsed = ConfigParser.parseDimOpts(opts, dims, enumMap);

        expect(parsed.enumType).to.be.equal(enumMap["enumtype:5"].name);
    });

    it("should fail to parse enumType with dimension", () => {

        let opts: DimensionStrOptions =  {
            name: "dim:day",
            dataType: "enumtype:4",
            parent: "dim:0",
            relation: "day"
        };
        let dims: Dimension[] = [
            new Dimension({name: "dim:0", dataType: DataType.INTEGER})
        ];
        let enumMap: {[key: string]: EnumType} = {
            "enumtype:5" : new EnumType({name: "enumtype:5", values: ["nope", "test"]})
        };
        let error: boolean = false;
        try {
            ConfigParser.parseDimOpts(opts, dims, enumMap);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] DataType: '" + opts.dataType + "' does not exist on Dimension");
        }
        expect(error).to.be.true;

    });

    it("should fail to parse dataType with Metrics", () => {
        let met: MetricStrOptions;
        met = {
            name: "met:0",
            aggregation: "sum",
            dataType: "string",
            description: "just for test"
        };
        let error: boolean = false;
        try {
            ConfigParser.parseMetOpts(met);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] DataType: '" + met.dataType + "' does not exist on Metric");
        }
        expect(error).to.be.true;

    });
    it("should fail to parse enumType with Source", () => {
        let sourc: SourceStrOptions;
        sourc = {
            name: "source_10",
            description: "source used for test only",
            fields: [
                    {name: "fields:0",
                    description: "first entry",
                    dataType: "string"},
                    {name: "fields:1",
                    description: "second entry",
                    dataType: "enumtype",
                    enumType: "enumtype:10"}
            ]
        };
        let enumMap: {[key: string]: EnumType} = {
            "enumtype:5" : new EnumType({name: "enumtype:5", values: ["nope", "test"]})
        };
        let error: boolean = false;
        try {
            ConfigParser.parseSourceOpts(sourc, enumMap);
        }
        catch (e) {
            error = true;
            expect(e.message).to.be
            .equal("[Parsing error] EnumType: '" + sourc.fields[1].enumType + "' does not exist on Source");
        }
        expect(error).to.be.true;
    });

});
